//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Views/SampleDesigner/SampleToolBoxTreeWidget.cpp
//! @brief     Implements class SampleToolBoxTreeWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Views/SampleDesigner/SampleToolBoxTreeWidget.h"
#include "Base/Utils/Assert.h"
#include "GUI/Views/SampleDesigner/SampleToolBoxGroupView.h"

#include <QAction>
#include <QActionGroup>
#include <QApplication>
#include <QContextMenuEvent>
#include <QHeaderView>
#include <QItemDelegate>
#include <QMenu>
#include <QPainter>
#include <QTreeWidgetItem>

namespace {

//! Draws the group title items as buttons with an arrow to show "collapsed/expanded".
//! This work is based on qdesigner_internal::SheetDelegate
class GroupTitleDelegate : public QItemDelegate {
public:
    GroupTitleDelegate(QTreeView* view, QWidget* parent) : QItemDelegate(parent), m_view(view) {}

    virtual void paint(QPainter* painter, const QStyleOptionViewItem& option,
                       const QModelIndex& index) const override
    {
        const QAbstractItemModel* model = index.model();
        const bool isGroupTitleItem = !model->parent(index).isValid();

        if (!isGroupTitleItem) {
            QItemDelegate::paint(painter, option, index);
            return;
        }

        painter->save();
        QColor buttonColor(230, 230, 230);
        const QBrush buttonBrush = option.palette.button();
        if (!buttonBrush.gradient() && buttonBrush.texture().isNull())
            buttonColor = buttonBrush.color();
        QColor outlineColor = buttonColor.darker(150);
        QColor highlightColor = buttonColor.lighter(130);

        // Only draw topline if the previous item is expanded
        const QModelIndex previousIndex = model->index(index.row() - 1, index.column());
        const bool drawTopline = (index.row() > 0 && m_view->isExpanded(previousIndex));
        const int highlightOffset = drawTopline ? 1 : 0;

        QLinearGradient gradient(option.rect.topLeft(), option.rect.bottomLeft());
        gradient.setColorAt(0, buttonColor.lighter(102));
        gradient.setColorAt(1, buttonColor.darker(106));

        painter->setPen(Qt::NoPen);
        painter->setBrush(gradient);
        painter->drawRect(option.rect);
        painter->setPen(highlightColor);
        painter->drawLine(option.rect.topLeft() + QPoint(0, highlightOffset),
                          option.rect.topRight() + QPoint(0, highlightOffset));
        painter->setPen(outlineColor);
        if (drawTopline)
            painter->drawLine(option.rect.topLeft(), option.rect.topRight());
        painter->drawLine(option.rect.bottomLeft(), option.rect.bottomRight());
        painter->restore();

        QStyleOption branchOption;
        static const int i = 9;
        const QRect r = option.rect;
        branchOption.rect = QRect(r.left() + i / 2, r.top() + (r.height() - i) / 2, i, i);
        branchOption.palette = option.palette;
        branchOption.state = QStyle::State_Children;

        if (m_view->isExpanded(index))
            branchOption.state |= QStyle::State_Open;

        m_view->style()->drawPrimitive(QStyle::PE_IndicatorBranch, &branchOption, painter, m_view);

        // draw text
        const QRect textrect =
            QRect(r.left() + i * 2, r.top(), r.width() - ((5 * i) / 2), r.height());
        const QString text = elidedText(option.fontMetrics, textrect.width(), Qt::ElideMiddle,
                                        model->data(index, Qt::DisplayRole).toString());
        m_view->style()->drawItemText(painter, textrect, Qt::AlignCenter, option.palette,
                                      m_view->isEnabled(), text);
    }

    virtual QSize sizeHint(const QStyleOptionViewItem& opt, const QModelIndex& index) const override
    {
        return QItemDelegate::sizeHint(opt, index) + QSize(2, 2);
    }

private:
    QTreeView* m_view;
};

} // namespace

SampleToolBoxTreeWidget::SampleToolBoxTreeWidget(QWidget* parent)
    : QTreeWidget(parent), m_iconMode(false)
{
    setFocusPolicy(Qt::NoFocus);
    setIndentation(0);
    setRootIsDecorated(false);
    setColumnCount(1);
    header()->hide();
#if QT_VERSION >= 0x050000
    header()->setSectionResizeMode(QHeaderView::Stretch);
#endif
    setTextElideMode(Qt::ElideMiddle);
    setVerticalScrollMode(ScrollPerPixel);

    setItemDelegate(new GroupTitleDelegate(this, this));

    connect(this, &QTreeWidget::itemPressed, this, &SampleToolBoxTreeWidget::handleMousePress);
}

SampleToolBoxGroupView* SampleToolBoxTreeWidget::groupViewAt(int idx) const
{
    SampleToolBoxGroupView* rc = nullptr;
    if (QTreeWidgetItem* groupItem = topLevelItem(idx))
        if (QTreeWidgetItem* embedItem = groupItem->child(0))
            rc = qobject_cast<SampleToolBoxGroupView*>(itemWidget(embedItem, 0));
    ASSERT(rc);
    return rc;
}

void SampleToolBoxTreeWidget::handleMousePress(QTreeWidgetItem* item)
{
    if (item == nullptr)
        return;

    if (QApplication::mouseButtons() != Qt::LeftButton)
        return;

    if (item->parent() == nullptr) {
#if QT_VERSION >= QT_VERSION_CHECK(5, 13, 0)
        item->setExpanded(!item->isExpanded());
#else
        setItemExpanded(item, !isItemExpanded(item));
#endif
        return;
    }
}

void SampleToolBoxTreeWidget::addGroupView(QTreeWidgetItem* parent, bool iconMode)
{
    QTreeWidgetItem* embed_item = new QTreeWidgetItem(parent);
    embed_item->setFlags(Qt::ItemIsEnabled);
    SampleToolBoxGroupView* view = new SampleToolBoxGroupView(this);
    view->setViewMode(iconMode ? QListView::IconMode : QListView::ListMode);
    connect(view, &SampleToolBoxGroupView::pressed, this, &SampleToolBoxTreeWidget::pressed);
    setItemWidget(embed_item, 0, view);
}

void SampleToolBoxTreeWidget::adjustSubListSize(QTreeWidgetItem* groupItem)
{
    QTreeWidgetItem* embedItem = groupItem->child(0);
    if (embedItem == nullptr)
        return;

    SampleToolBoxGroupView* list_widget =
        static_cast<SampleToolBoxGroupView*>(itemWidget(embedItem, 0));
    list_widget->setFixedWidth(header()->width());
    list_widget->doItemsLayout();
    const int height = qMax(list_widget->contentsSize().height(), 1);
    list_widget->setFixedHeight(height);
    embedItem->setSizeHint(0, QSize(-1, height - 1));
}

void SampleToolBoxTreeWidget::slotListMode()
{
    m_iconMode = false;
    updateViewMode();
}

void SampleToolBoxTreeWidget::slotIconMode()
{
    m_iconMode = true;
    updateViewMode();
}

void SampleToolBoxTreeWidget::updateViewMode()
{
    if (const int numTopLevels = topLevelItemCount()) {
        for (int i = numTopLevels - 1; i >= 0; --i) {
            const QListView::ViewMode viewMode =
                m_iconMode ? QListView::IconMode : QListView::ListMode;
            SampleToolBoxGroupView* view = groupViewAt(i);
            if (viewMode != view->viewMode()) {
                view->setViewMode(viewMode);
                adjustSubListSize(topLevelItem(i));
            }
        }
    }
    updateGeometries();
}

void SampleToolBoxTreeWidget::resizeEvent(QResizeEvent* e)
{
    QTreeWidget::resizeEvent(e);
    if (const int numTopLevels = topLevelItemCount()) {
        for (int i = numTopLevels - 1; i >= 0; --i)
            adjustSubListSize(topLevelItem(i));
    }
}

void SampleToolBoxTreeWidget::contextMenuEvent(QContextMenuEvent* e)
{
    QTreeWidgetItem* item = itemAt(e->pos());

    QMenu menu;
    menu.addAction("Expand all", this, SLOT(expandAll()));
    menu.addAction("Collapse all", this, SLOT(collapseAll()));
    menu.addSeparator();

    QAction* listModeAction = menu.addAction("List View");
    QAction* iconModeAction = menu.addAction("Icon View");
    listModeAction->setCheckable(true);
    iconModeAction->setCheckable(true);
    QActionGroup* viewModeGroup = new QActionGroup(&menu);
    viewModeGroup->addAction(listModeAction);
    viewModeGroup->addAction(iconModeAction);
    if (m_iconMode)
        iconModeAction->setChecked(true);
    else
        listModeAction->setChecked(true);
    connect(listModeAction, &QAction::triggered, this, &SampleToolBoxTreeWidget::slotListMode);
    connect(iconModeAction, &QAction::triggered, this, &SampleToolBoxTreeWidget::slotIconMode);

    e->accept();
    menu.exec(mapToGlobal(e->pos()));
}

void SampleToolBoxTreeWidget::filter(const QString& f)
{
    const bool empty = f.isEmpty();
    QRegExp re = empty ? QRegExp() : QRegExp(f, Qt::CaseInsensitive, QRegExp::FixedString);
    const int numTopLevels = topLevelItemCount();
    bool changed = false;
    for (int i = 0; i < numTopLevels; i++) {
        QTreeWidgetItem* tl = topLevelItem(i);
        SampleToolBoxGroupView* groupView = groupViewAt(i);
        // Anything changed? -> Enable the group
        const int oldCount = groupView->count(SampleToolBoxGroupView::FILTERED);
        groupView->filter(re);
        const int newCount = groupView->count(SampleToolBoxGroupView::FILTERED);
        if (oldCount != newCount) {
            changed = true;
            const bool groupEnabled = newCount > 0 || empty;
            if (groupEnabled) {
                groupView->adjustSize();
                adjustSubListSize(tl);
            }
            setRowHidden(i, QModelIndex(), !groupEnabled);
        }
    }
    if (changed)
        updateGeometries();
}

void SampleToolBoxTreeWidget::addGroup(const QString& title)
{
    QTreeWidgetItem* groupItem = new QTreeWidgetItem();
    groupItem->setText(0, title);
    addTopLevelItem(groupItem);
#if QT_VERSION >= QT_VERSION_CHECK(5, 13, 0)
    groupItem->setExpanded(true);
#else
    setItemExpanded(groupItem, true);
#endif

    addGroupView(groupItem, m_iconMode);
    adjustSubListSize(groupItem);
}

void SampleToolBoxTreeWidget::addEntry(const QString& title, const QIcon& icon,
                                       const QString& className, const QString& tooltip)
{
    const auto groupIndex = topLevelItemCount() - 1;
    groupViewAt(groupIndex)->addEntry(title, icon, className, tooltip);
    adjustSubListSize(topLevelItem(groupIndex));
}
